import collections
import json
from typing import Any, Dict, List, Optional

from . import errors as E
from . import constants as C
from .color import color
from .logger import logger

import yaml


class ConfigBase(collections.UserDict):
    def __init__(self, config: Dict[str, str], path: Optional[str] = None) -> None:
        super().__init__(config)
        self.validate_config()
        self.path: Optional[str] = path

    @classmethod
    def from_yaml_filepath(cls, path: str):
        with open(path) as raw_config:
            return cls(yaml.safe_load(raw_config), path)

    def display_options(self) -> None:
        cls = type(self)
        names = [name for name in dir(cls) if isinstance(getattr(cls, name), property)]
        for name in sorted(names):
            value = json.dumps(getattr(self, name))
            logger.info(f"{name}: {color.config(value)}")

    def validate_config(self) -> None:
        if "targets" not in self.data:
            raise E.InvalidConfigError("targets section missing")

        if not isinstance(self.data["targets"], dict):
            raise E.InvalidConfigError("dict expected for targets section")

        if not self.data["targets"]:
            raise E.InvalidConfigError("no targets defined")

        for target in self.data["targets"]:
            self.validate_target(target)

    def validate_target(self, target: str) -> None:
        config = self.data["targets"][target]

        if not isinstance(config, dict):
            raise E.InvalidConfigError(f"{target}: dict expected")

        if "dockerfile" not in config or not isinstance(config["dockerfile"], str):
            raise E.InvalidConfigError(
                f"{target}: dockerfile required in target config"
            )

        self.validate_target_config(target, config)

    def validate_target_config(self, target: str, config: Dict[str, Any]) -> None:
        if not isinstance(config.get("depends_on", []), list):
            raise E.InvalidConfigError(f"{target}: list expected for depends_on")

        if not isinstance(config.get("watch_files", []), list):
            raise E.InvalidConfigError(f"{target}: list expected for watch_files")

        if not isinstance(config.get("build_options", {}), dict):
            raise E.InvalidConfigError(f"{target}: dict expected for build_options")


class Config(ConfigBase):
    def get_target(self, name: str) -> Dict[str, Any]:
        target = self.data.get("targets", {}).get(name)

        if target is None:
            raise E.TargetNotFoundError(f"Target [{name}] not found in build config")

        for key in ["depends_on", "watch_files"]:
            if key not in target:
                target[key] = []

        if "build_options" not in target:
            target["build_options"] = {}

        return target

    def get_option(self, name: str, default: Any) -> Any:
        return self.data.get("options", {}).get(name, default)

    @property
    def allowed_pull_errors(self) -> List[str]:
        return self.get_option("allowed_pull_errors", C.DEFAULT_ALLOWED_PULL_ERRORS)

    @property
    def buildarg_prefix(self) -> str:
        return self.get_option("buildarg_prefix", C.DEFAULT_BUILDARG_PREFIX)

    @property
    def build_on_validate_error(self) -> str:
        return self.get_option(
            "build_on_validate_error", C.DEFAULT_BUILD_ON_VALIDATE_ERROR
        )

    @property
    def color_style(self) -> Dict[str, int]:
        return self.get_option("color_style", C.DEFAULT_COLOR_STYLE)

    @property
    def digest_label(self) -> str:
        return self.get_option("digest_label", C.DEFAULT_DIGEST_LABEL)

    @property
    def hash_algorithm(self) -> str:
        return self.get_option("hash_algorithm", C.DEFAULT_HASH_ALGORITHM)

    @property
    def hash_chunk_size(self) -> int:
        return self.get_option("hash_chunk_size", C.DEFAULT_HASH_CHUNK_SIZE)

    @property
    def ignore_files(self) -> List[str]:
        return self.get_option("ignore_files", C.DEFAULT_IGNORE_FILES)

    @property
    def prune_after_build(self) -> bool:
        return self.get_option("prune_after_build", C.DEFAULT_PRUNE_AFTER_BUILD)

    @property
    def short_digest_size(self) -> int:
        return self.get_option("short_digest_size", C.DEFAULT_SHORT_DIGEST_SIZE)

    @property
    def target_label(self) -> str:
        return self.get_option("target_label", C.DEFAULT_TARGET_LABEL)

    @property
    def validate_image(self) -> bool:
        return self.get_option("validate_image", C.DEFAULT_VALIDATE_IMAGE)



import os
import sys

import pytest


LIB_PATH = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "lib"))
FIXTURES_PATH = os.path.abspath(os.path.join(os.path.dirname(__file__), "fixtures"))
sys.path.insert(0, LIB_PATH)


@pytest.fixture
def prefab_yaml_path():
    return os.path.join(FIXTURES_PATH, "prefab.yml")


@pytest.fixture
def chdir_fixtures(monkeypatch):
    monkeypatch.chdir(FIXTURES_PATH)



from typing import Any, Dict, List, Optional


DEFAULT_ALLOWED_PULL_ERRORS: List[str] = [
    "ImageAccessError",
    "ImageNotFoundError",
    "ImageValidationError",
]
DEFAULT_BUILD_OPTIONS: Dict[str, Any] = {
    "forcerm": True,
    "path": ".",
    "rm": True,
    "squash": False,
}
DEFAULT_BUILDARG_PREFIX = "PREFAB_"
DEFAULT_BUILD_ON_VALIDATE_ERROR = True
DEFAULT_COLOR_STYLE: Dict[str, Optional[int]] = {
    "config": 2,
    "elapsed": 11,
    "error": 1,
    "header": None,
    "image": 3,
    "target": 6,
    "warning": 1,
}
DEFAULT_CONFIG_FILE = "prefab.yml"
DEFAULT_DIGEST_LABEL = "prefab.digest"
DEFAULT_ENVFILES: Dict[str, str] = {
    "REGISTRY_AUTH": "/auth.json",
}
DEFAULT_HASH_ALGORITHM = "sha256"
DEFAULT_HASH_CHUNK_SIZE = 65535
DEFAULT_IGNORE_FILES: List[str] = []
DEFAULT_PRUNE_AFTER_BUILD = True
DEFAULT_SHORT_DIGEST_SIZE = 12
DEFAULT_TARGET_LABEL = "prefab.target"
DEFAULT_VALIDATE_IMAGE = True



from sanic import Sanic, response
import html5_parser

app = Sanic(name="prefab-example")


@app.post("/")
async def html_to_text(request):
    tree = html5_parser.parse(request.body)
    return response.text(tree.xpath("normalize-space(//*)"))


if __name__ == "__main__":
    app.run(host="0.0.0.0", port=8000)



import pytest

from prefab import errors as E
from prefab.color import color
from prefab.config import Config
from prefab.image import FakeImage, ImageFactory, ImageGraph


def create_graph(targets):
    config = Config({"targets": targets}, "prefab.yaml")
    image_factory = ImageFactory(
        config, "repo", tags=dict(), image_constructor=FakeImage
    )

    return ImageGraph(config, image_factory)


@pytest.fixture(autouse=True)
def disable_color(monkeypatch):
    monkeypatch.setattr(color, "enabled", False)


@pytest.fixture
def a_graph(chdir_fixtures):
    targets = {
        "a": {
            "dockerfile": "Dockerfile.a",
        }
    }

    return create_graph(targets)


@pytest.fixture
def abc_graph(chdir_fixtures):
    targets = {
        "a": {
            "dockerfile": "Dockerfile.a",
            "depends_on": ["b"],
        },
        "b": {
            "dockerfile": "Dockerfile.b",
            "depends_on": ["c"],
        },
        "c": {
            "dockerfile": "Dockerfile.c",
        },
    }

    return create_graph(targets)


def resolve_graph(targets, target):
    config = Config({"targets": targets})
    image_factory = ImageFactory(
        config, "repo", tags=dict(), image_constructor=FakeImage
    )
    image_factory.digests = {target: target for target in targets}
    image_graph = ImageGraph(config, image_factory)

    return image_graph.resolve_target_dependencies(target)


def test_single_target():
    targets = {
        "a": {
            "dockerfile": "Dockerfile.a",
        }
    }

    assert resolve_graph(targets, "a") == ["a"]


def test_line_graph():
    targets = {
        "a": {
            "dockerfile": "Dockerfile.a",
            "depends_on": ["b"],
        },
        "b": {
            "dockerfile": "Dockerfile.b",
            "depends_on": ["c"],
        },
        "c": {
            "dockerfile": "Dockerfile.c",
            "depends_on": ["d"],
        },
        "d": {
            "dockerfile": "Dockerfile.d",
        },
    }

    assert resolve_graph(targets, "a") == ["d", "c", "b", "a"]


def test_diamond_graph():
    targets = {
        "a": {
            "dockerfile": "Dockerfile.a",
            "depends_on": ["b", "c"],
        },
        "b": {
            "dockerfile": "Dockerfile.b",
            "depends_on": ["d"],
        },
        "c": {
            "dockerfile": "Dockerfile.c",
            "depends_on": ["d"],
        },
        "d": {
            "dockerfile": "Dockerfile.d",
        },
    }

    assert resolve_graph(targets, "a") == ["d", "b", "c", "a"]


def test_fake_app_test_graph():
    targets = {
        "test": {
            "dockerfile": "Dockerfile.test",
            "depends_on": ["app", "test_pkgs"],
        },
        "app": {
            "dockerfile": "Dockerfile.app",
            "depends_on": ["base", "app_pkgs"],
        },
        "app_pkgs": {
            "dockerfile": "Dockerfile.app_pkgs",
            "depends_on": ["tools"],
        },
        "test_pkgs": {
            "dockerfile": "Dockerfile.test_pkgs",
            "depends_on": ["tools"],
        },
        "tools": {
            "dockerfile": "Dockerfile.tools",
            "depends_on": ["base"],
        },
        "base": {
            "dockerfile": "Dockerfile.base",
        },
    }

    assert resolve_graph(targets, "test") == [
        "base",
        "tools",
        "app_pkgs",
        "app",
        "test_pkgs",
        "test",
    ]


def test_self_loop():
    targets = {
        "a": {
            "dockerfile": "Dockerfile.a",
            "depends_on": ["a"],
        }
    }

    with pytest.raises(E.TargetCyclicError):
        resolve_graph(targets, "a")


def test_short_loop():
    targets = {
        "a": {
            "dockerfile": "Dockerfile.a",
            "depends_on": ["b"],
        },
        "b": {
            "dockerfile": "Dockerfile.b",
            "depends_on": ["a"],
        },
    }

    for target in targets:
        with pytest.raises(E.TargetCyclicError):
            resolve_graph(targets, target)


def test_long_loop():
    targets = {
        "a": {
            "dockerfile": "Dockerfile.a",
            "depends_on": ["b"],
        },
        "b": {
            "dockerfile": "Dockerfile.b",
            "depends_on": ["c"],
        },
        "c": {
            "dockerfile": "Dockerfile.c",
            "depends_on": ["d"],
        },
        "d": {
            "dockerfile": "Dockerfile.d",
            "depends_on": ["a"],
        },
    }

    for target in targets:
        with pytest.raises(E.TargetCyclicError):
            resolve_graph(targets, target)


def test_build_on_validation_error(caplog, a_graph):
    a_graph.images = {
        "a": FakeImage(
            "repo",
            "tag",
            build_options=dict(),
            loaded=True,
            validate=E.ImageValidationError,
        )
    }
    a_graph.build(["a"])

    assert caplog.messages[2:] == [
        "repo:tag Image loaded",
        "repo:tag ImageValidationError: ",
        "repo:tag build_on_validate_error enabled, continuing...",
        "repo:tag Trying build...",
        "repo:tag build_options {}",
        "repo:tag Build succeeded",
    ]


def test_build_on_allowed_pull_error(caplog, a_graph):
    a_graph.images = {
        "a": FakeImage(
            "repo",
            "tag",
            build_options=dict(),
            loaded=False,
            pull=E.ImageNotFoundError,
        )
    }
    a_graph.build(["a"])

    assert caplog.messages[2:] == [
        "repo:tag Image not loaded",
        "repo:tag Trying pull...",
        "repo:tag ImageNotFoundError: ",
        "repo:tag ImageNotFoundError in allowed_pull_errors, continuing...",
        "repo:tag Trying build...",
        "repo:tag build_options {}",
        "repo:tag Build succeeded",
    ]


def test_target_a_image_pull_stops_graph_traversal(caplog, abc_graph):
    abc_graph.build(["a"])

    assert caplog.messages[-4:] == [
        "\nBuilding [a] target with dependencies: [b], [c]",
        "[a] repo:33cb9338a826 Image not loaded",
        "[a] repo:33cb9338a826 Trying pull...",
        "[a] repo:33cb9338a826 Image validated",
    ]


def test_target_b_image_pull_stops_graph_traversal(caplog, abc_graph):
    abc_graph.resolve_build_targets(["a"])
    abc_graph.images["a"].raise_on.pull = E.ImageNotFoundError
    abc_graph.build(["a"])

    assert caplog.messages[-25:-21] == [
        "\nBuilding [a] target with dependencies: [b], [c]",
        "[a] repo:33cb9338a826 Image not loaded",
        "[a] repo:33cb9338a826 Trying pull...",
        "[a] repo:33cb9338a826 ImageNotFoundError: ",
    ]
    assert caplog.messages[-20:-16] == [
        "[b] repo:a413f339464a Image not loaded",
        "[b] repo:a413f339464a Trying pull...",
        "[b] repo:a413f339464a Image validated",
        "[a] repo:33cb9338a826 Trying build...",
    ]
    assert caplog.messages[-1] == "[a] repo:33cb9338a826 Build succeeded"



import logging
from typing import Any, MutableMapping, Tuple

from .color import color


logger = logging.getLogger("prefab")
logger.handlers = [logging.StreamHandler()]
logger.setLevel(logging.INFO)


class TargetLoggerAdapter(logging.LoggerAdapter):
    def process(
        self, message: str, kwargs: MutableMapping[str, Any]
    ) -> Tuple[str, MutableMapping[str, Any]]:
        prefix = color.target(f"[{self.extra.get('target', '<none>')}]")
        return f"{prefix} {message}", kwargs



import base64
import os
import tempfile

import pytest

from prefab import envfile
from prefab.color import color


ENV_KEY = "REGISTRY_AUTH"


@pytest.fixture(autouse=True)
def disable_color(monkeypatch):
    monkeypatch.setattr(color, "enabled", False)


@pytest.fixture
def auth_env_key(monkeypatch):
    monkeypatch.setenv(ENV_KEY, base64.b64encode(b'{"auths":{}}\n').decode())


@pytest.fixture
def bad_env_key(monkeypatch):
    monkeypatch.setenv(ENV_KEY, "A" * 7)


@pytest.fixture
def empty_env_key(monkeypatch):
    monkeypatch.setenv(ENV_KEY, "")


@pytest.fixture
def dest_path():
    dest = tempfile.NamedTemporaryFile()
    dest.close()
    yield dest.name

    if os.path.exists(dest.name):
        os.unlink(dest.name)


@pytest.fixture
def dockerenv_path():
    dockerenv = tempfile.NamedTemporaryFile()
    yield dockerenv.name


@pytest.fixture
def no_dockerenv_path():
    dockerenv = tempfile.NamedTemporaryFile()
    dockerenv.close()
    yield dockerenv.name


def test_write_simple(auth_env_key, dest_path, dockerenv_path):
    envfile.write(ENV_KEY, dest_path, dockerenv_path)

    with open(dest_path) as dest:
        assert dest.read() == '{"auths":{}}\n'
    assert os.path.exists(dockerenv_path)


def test_no_write_if_dockernv_does_not_exist(
    auth_env_key, dest_path, no_dockerenv_path
):
    envfile.write(ENV_KEY, dest_path, no_dockerenv_path)

    assert not os.path.exists(dest_path)
    assert not os.path.exists(no_dockerenv_path)


def test_no_write_if_env_var_empty(empty_env_key, dest_path, dockerenv_path, caplog):
    envfile.write(ENV_KEY, dest_path, dockerenv_path)

    assert not os.path.exists(dest_path)
    assert os.path.exists(dockerenv_path)
    assert f"Not writing empty env REGISTRY_AUTH to {dest_path}" in caplog.text


def test_write_handles_decode_error(bad_env_key, dest_path, dockerenv_path, caplog):
    envfile.write(ENV_KEY, dest_path, dockerenv_path)

    assert not os.path.exists(dest_path)
    assert (
        f"Not writing env REGISTRY_AUTH to {dest_path}, base64 decode failed: Incorrect padding"
        in caplog.text
    )


def test_write_handles_io_error(auth_env_key, dockerenv_path, caplog):
    envfile.write(ENV_KEY, "/sys/is_a_ro_fs", dockerenv_path)

    assert "Failed to write REGISTRY_AUTH to /sys/is_a_ro_fs" in caplog.text



import sys

import pytest

from prefab.cli import main


def test_dry_run(caplog, monkeypatch, prefab_yaml_path, chdir_fixtures):
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "calling_path",
            "--dry-run",
            "--monochrome",
            "--config",
            prefab_yaml_path,
            "--repo",
            "lexsca/prefab",
            "--target",
            "a:a",
            "--push",
            "c",
            "b",
            "a",
        ],
    )

    with pytest.raises(SystemExit) as system_exit:
        main()

    target_digest_log_message = (
        "[a] target_digest sha256:465e3e6c70a73d0e704a"
        "109fde6b869b06db9a2afd2d0faaf76c174a1d3fd58c"
    )

    assert system_exit.value.code == 0
    assert caplog.records[-11].message == "[a] lexsca/prefab:a Build succeeded"
    assert caplog.records[-3].message == "[a] lexsca/prefab:a Pushed"
    assert target_digest_log_message in caplog.text


def test_tag_duplicates_disallowed(monkeypatch, capsys, prefab_yaml_path):
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "calling_path",
            "--config",
            prefab_yaml_path,
            "--repo",
            "lexsca/prefab",
            "--target",
            "a:app",
            "b:app",
        ],
    )

    with pytest.raises(SystemExit) as system_exit:
        main()

    assert capsys.readouterr().err.splitlines()[-1] == "Duplicate tag: app"
    assert system_exit.value.code == 2


def test_bad_arg_exits_non_zero(monkeypatch):
    monkeypatch.setattr(sys, "argv", ["calling_path", "--not-valid-arg"])

    with pytest.raises(SystemExit) as system_exit:
        main()

    assert system_exit.value.code == 2


def test_exception_exits_non_zero(monkeypatch):
    monkeypatch.setattr(sys, "argv", ["calling_path", "-r", "repo", "-t", "tag"])

    with pytest.raises(SystemExit) as system_exit:
        main()

    assert system_exit.value.code == 1



import pytest

from prefab.color import color


@pytest.fixture(autouse=True)
def enable_color(monkeypatch):
    monkeypatch.setattr(color, "enabled", True)


@pytest.fixture
def header_color_red(monkeypatch):
    monkeypatch.setitem(color.style, "header", 1)


@pytest.fixture
def header_color_none(monkeypatch):
    monkeypatch.setitem(color.style, "header", None)


def test_header_no_color(header_color_none):
    result = color.header("decline-voguish-misty-pause")

    assert result == "decline-voguish-misty-pause"


def test_header_red_color(header_color_red):
    result = color.header("cryptic-throve-lore-edict")

    assert result == "\x1b[38;5;1mcryptic-throve-lore-edict\x1b[0m"



import functools
import hashlib
import os
import stat
from typing import Any, Callable, Dict, List

from .. import constants as C
from .. import errors as E
from ..config import Config
from ..logger import logger, TargetLoggerAdapter
from ..walk import walk
from .docker import DockerImage


class ImageFactory:
    def __init__(
        self,
        config: Config,
        repo: str,
        tags: Dict[str, str],
        image_constructor: Callable = DockerImage,
    ) -> None:
        self.config: Config = config
        self.repo: str = repo
        self.tags: Dict[str, str] = tags
        self.image_constructor: Callable = image_constructor
        self.digests: Dict[str, str] = dict()

    def __call__(self, target: str) -> DockerImage:
        return self.image_constructor(
            repo=self.repo,
            tag=self.get_target_tag(target),
            build_options=self.get_target_build_options(target),
            logger=self.get_target_logger(target),
        )

    def get_target_logger(self, target: str) -> TargetLoggerAdapter:
        return TargetLoggerAdapter(logger, extra={"target": target})

    def get_hasher(self):
        if self.config.hash_algorithm not in hashlib.algorithms_available:
            raise E.HashAlgorithmNotFound(
                f"{self.config.hash_algorithm} not found in hashlib"
            )

        return getattr(hashlib, self.config.hash_algorithm)()

    def get_file_digest(self, path: str) -> str:
        hasher = self.get_hasher()
        chunk_size = self.config.hash_chunk_size

        with open(path, "rb") as file_data:
            chunker = functools.partial(file_data.read, chunk_size)
            for chunk in iter(chunker, b""):
                hasher.update(chunk)

        return hasher.hexdigest()

    def _get_target_watch_files(self, target: str) -> List[str]:
        target_config = self.config.get_target(target)
        watch_files = [target_config["dockerfile"]]

        for path in target_config["watch_files"]:
            if stat.S_ISDIR(os.stat(path).st_mode):
                watch_files.extend(walk(path, self.config.ignore_files))
            else:
                watch_files.append(path)

        return watch_files

    def get_target_salt(self, target: str) -> str:
        hasher = self.get_hasher()
        hasher.update(target.encode())

        return hasher.hexdigest()

    def _get_target_digest(self, target: str) -> str:
        hasher = self.get_hasher()
        target_logger = self.get_target_logger(target)
        target_config = self.config.get_target(target)

        target_salt = self.get_target_salt(target)
        target_logger.info(f"target_salt {hasher.name}:{target_salt}")
        hasher.update(target_salt.encode())

        for path in self._get_target_watch_files(target):
            digest = self.get_file_digest(path)
            hasher.update(digest.encode())
            target_logger.info(f"file_digest {path} {hasher.name}:{digest}")

        for dependency in target_config["depends_on"]:
            hasher.update(self.digests[dependency].encode())
            target_logger.info(
                f"depends_on {dependency} {hasher.name}:{self.digests[dependency]}"
            )

        digest = hasher.hexdigest()
        target_logger.info(f"target_digest {hasher.name}:{digest}")

        return digest

    def get_target_digest(self, target: str) -> str:
        if target not in self.digests:
            self.digests[target] = self._get_target_digest(target)

        return self.digests[target]

    def get_target_tag(self, target: str) -> str:
        if target not in self.tags:
            digest = self.get_target_digest(target)
            self.tags[target] = digest[: self.config.short_digest_size]

        return self.tags[target]

    def get_target_labels(self, target: str) -> Dict[str, str]:
        digest = self.get_target_digest(target)
        hasher = self.get_hasher()

        return {
            self.config.target_label: target,
            self.config.digest_label: f"{hasher.name}:{digest}",
        }

    def get_target_buildargs(self, target: str) -> Dict[str, str]:
        buildargs = {}

        for dependency in self.config.get_target(target)["depends_on"]:
            arg = f"{self.config.buildarg_prefix}{dependency}".upper().replace("-", "_")
            value = f"{self.repo}:{self.get_target_tag(dependency)}"
            buildargs[arg] = value

        return buildargs

    def get_target_build_options(self, target: str) -> Dict[str, Any]:
        target_config = self.config.get_target(target)
        build_options: Dict[str, Any] = {
            **C.DEFAULT_BUILD_OPTIONS,
            **{
                "dockerfile": target_config["dockerfile"],
                "labels": self.get_target_labels(target),
                "buildargs": self.get_target_buildargs(target),
                "tag": f"{self.repo}:{self.get_target_tag(target)}",
            },
        }

        for key, value in target_config.get("build_options", {}).items():
            if key in {"labels", "buildargs"}:
                build_options[key].update(value)
            else:
                build_options[key] = value

        return build_options



from typing import Dict, Optional

from . import constants as C


class Color:
    reset: str = "\x1b[0m"

    def __init__(
        self,
        style: Dict[str, Optional[int]] = C.DEFAULT_COLOR_STYLE,
        enabled: bool = True,
    ) -> None:
        self.enabled: bool = enabled
        self.style: Dict[str, Optional[int]] = style.copy()

    def _colorize(self, style_name: str, text: str) -> str:
        # https://en.wikipedia.org/wiki/ANSI_escape_code#8-bit
        color_code = self.style.get(style_name)
        color_seq = f"\x1b[38;5;{color_code}m"
        colorize = self.enabled and color_code is not None

        return f"{color_seq}{text}{self.reset}" if colorize else text

    def config(self, text: str) -> str:
        return self._colorize("config", text)

    def elapsed(self, text: str) -> str:
        return self._colorize("elapsed", text)

    def error(self, text: str) -> str:
        return self._colorize("error", text)

    def header(self, text: str) -> str:
        return self._colorize("header", text)

    def image(self, text: str) -> str:
        return self._colorize("image", text)

    def target(self, text: str) -> str:
        return self._colorize("target", text)

    def warning(self, text: str) -> str:
        return self._colorize("warning", text)


color = Color()



import io
import os

import pytest

from prefab import constants as C
from prefab.image import DockerImage


@pytest.mark.skipif(not os.path.exists("/var/run/docker.sock"), reason="docker missing")
def test_pull_prefab_none():
    build_options = {"labels": {"prefab.target": "none"}}
    image = DockerImage(repo="lexsca/prefab", tag="none", build_options=build_options)
    image.pull()
    image.validate()
    image.remove()

    assert image.is_loaded


@pytest.mark.skipif(not os.path.exists("/var/run/docker.sock"), reason="docker missing")
def test_build_from_scratch():
    fileobj = io.BytesIO(b"FROM scratch")
    build_options = {
        **C.DEFAULT_BUILD_OPTIONS,
        **{
            "fileobj": fileobj,
            "decode": True,
            "labels": {"pacifist-kept": "chewy-uncouth"},
            "tag": "sunlit-sing:aspect-workbag",
            "squash": False,
        },
    }
    image = DockerImage(
        repo="sunlit-sing", tag="aspect-workbag", build_options=build_options
    )
    image.build()
    image.validate()
    image.remove()

    assert image.is_loaded


@pytest.mark.skipif(not os.path.exists("/var/run/docker.sock"), reason="docker missing")
def test_loaded_does_not_raise_ImageNotFoundError():
    image = DockerImage(repo="scratch", tag=None, build_options={})

    assert not image.is_loaded



from setuptools import setup

setup(
    name="container-prefab",
    description="Build container images faster ⚡️",
    long_description=open("README.rst").read(),
    author="Lex Scarisbrick",
    author_email="lex@scarisbrick.org",
    license="MIT",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Topic :: Software Development :: Build Tools",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    python_requires=">=3.7",
    url="https://github.com/lexsca/prefab.git",
    package_dir={"": "lib"},
    packages=["prefab", "prefab.image"],
    scripts=["bin/prefab"],
    install_requires=["docker", "pyyaml"],
    use_scm_version={
        "write_to": "lib/prefab/version.py",
        "write_to_template": "__version__ = '{version}'\n",
    },
)



import json
import logging
import os
from typing import Any, Dict, Generator, Optional, Union

import docker

from .. import errors as E
from ..color import color
from ..logger import logger


class DockerImage:
    def __init__(
        self,
        repo: str,
        tag: str,
        build_options: Dict[str, Any],
        docker_client: Optional[docker.client.DockerClient] = None,
        logger: Union[logging.Logger, logging.LoggerAdapter] = logger,
    ) -> None:
        self.repo: str = repo
        self.tag: str = tag
        self.build_options: Dict[str, Any] = build_options
        self.logger: Union[logging.Logger, logging.LoggerAdapter] = logger
        self._loaded: Optional[bool] = None
        self._docker_client = docker_client
        self.was_pulled: bool = False
        self.was_built: bool = False

    @property
    def docker_client(self):
        if self._docker_client is None:
            self._docker_client = docker.from_env(version="auto")

        return self._docker_client

    def _log_stream_decoder(
        self, chunker: Generator[bytes, None, None]
    ) -> Generator[Dict[str, Any], None, None]:
        for chunk in chunker:
            for line in chunk.splitlines():
                if not line:
                    continue
                try:
                    yield json.loads(line)
                except json.decoder.JSONDecodeError:
                    self.logger.warning(
                        color.warning(f"Skipping malformed log entry: {chunk!r}")
                    )

    def _format_transfer_message(self, log_entry: Dict[str, Any]) -> str:
        if "id" in log_entry:
            message = f"{log_entry.get('id')}: {log_entry.get('status')}"
        else:
            message = str(log_entry["status"])

        return message

    def _process_transfer_log_stream(
        self, log_stream: Generator[bytes, None, None]
    ) -> None:
        for log_entry in self._log_stream_decoder(log_stream):
            if "error" in log_entry:
                raise E.ImageAccessError(log_entry["error"])

            if "status" not in log_entry or log_entry.get("progressDetail"):
                continue

            message = self._format_transfer_message(log_entry)
            self.logger.info(message)

    def _process_build_log_stream(
        self, log_stream: Generator[bytes, None, None]
    ) -> None:
        for log_entry in self._log_stream_decoder(log_stream):
            for key in {"error", "message"}:
                if key in log_entry:
                    raise E.ImageBuildError(log_entry[key])

            message = log_entry.get("stream", "").rstrip()
            if message:
                self.logger.info(message)

    def _get_docker_image(self) -> docker.models.images.Image:
        docker_image = None

        for image in self.docker_client.images.list(name=self.repo):
            if self.name in image.tags:
                docker_image = image
                break

        if docker_image is None:
            raise E.ImageNotFoundError(f"{self.name} not found")

        return docker_image

    @property
    def name(self) -> str:
        return f"{self.repo}:{self.tag}"

    @property
    def is_loaded(self) -> bool:
        if self._loaded is None:
            try:
                self._get_docker_image()
                self._loaded = True
            except E.ImageNotFoundError:
                self._loaded = False

        return self._loaded

    def _pull_success(self) -> None:
        self._loaded = True
        self.was_pulled = True

    def pull(self) -> None:
        try:
            log_stream = self.docker_client.api.pull(
                repository=self.repo, tag=self.tag, stream=True, decode=False
            )
            self._process_transfer_log_stream(log_stream)
            self._pull_success()
        except docker.errors.NotFound as error:
            raise E.ImageNotFoundError(error.explanation)
        except docker.errors.APIError as error:
            raise E.ImageAccessError(error.explanation)

    def validate(self) -> None:
        docker_image = self._get_docker_image()

        for name, expected in self.build_options["labels"].items():
            result = docker_image.labels.get(name)
            if result != expected:
                raise E.ImageValidationError(
                    f'{self.name} label "{name}" '
                    f'expected value "{expected}", got "{result}"'
                )

    def _build_success(self) -> None:
        self.logger.info(f"{color.image(self.name)} Build succeeded")
        self.was_built = True
        self._loaded = True

    def build(self) -> None:
        try:
            build_options = {**self.build_options, **{"decode": False}}
            log_stream = self.docker_client.api.build(**build_options)
            self._process_build_log_stream(log_stream)
            self._build_success()
        except docker.errors.APIError as error:
            raise E.ImageBuildError(str(error))
        except Exception:
            if build_options.get("squash") and os.path.exists("/.dockerenv"):
                message = color.warning(
                    "The squash build_option may not be supported in dind "
                    "environments. Try disabling to mitigate this error:"
                )
                self.logger.warning(message)
            raise

    def prune(self) -> None:
        try:
            self.docker_client.api.prune_images(filters={"dangling": True})
        except Exception as error:
            self.logger.warning(color.warning(f"Prune dangling images failed: {error}"))

    def push(self) -> None:
        try:
            log_stream = self.docker_client.images.push(
                repository=self.repo, tag=self.tag, stream=True, decode=False
            )
            self._process_transfer_log_stream(log_stream)
        except docker.errors.APIError as error:
            raise E.ImagePushError(str(error))

    def remove(self, force: bool = False, noprune: bool = False) -> None:
        docker_image = self._get_docker_image()
        self.docker_client.images.remove(docker_image.id, force, noprune)
